from .download import download_zip, download_file
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_zip(xz_url + "Decals.zip",down_path)
download_zip(xz_url + "Textures.zip",down_path)
download_zip(xz_url + "dairin0d.zip",down_path)
download_file(xz_url + "default.blend",down_path)
download_file(xz_url + "Folder.png",down_path)

bl_info = {
    "name": "XDECAL™",
    "author": "moth3r, dairin0d",
    "version": (1, 0, 0),
    "blender": (3, 6, 0),
    "location": "3D View > XDECAL™ tab",
    "description": "Decal toolkit",
    "warning": "",
    "doc_url": "https://docs.moth3r.com",
    "tracker_url": "https://moth3r.com/info#contact",
    "category": "Object",
}

# =========================================================================== #

import math

import bpy
from mathutils import Vector, Matrix, Quaternion, Euler, Color

import importlib
import importlib.util

if "dairin0d" in locals():
    importlib.reload(icon_geometry)
    importlib.reload(common)
    importlib.reload(material_utils)
    importlib.reload(decal_utils)
    importlib.reload(decal_gizmos)
    importlib.reload(type_extensions)
    importlib.reload(palette)
    importlib.reload(palette_options)
    importlib.reload(palette_gizmo)
    importlib.reload(palette_ui)
    importlib.reload(template_material_menu)
    importlib.reload(decal_operators)
    importlib.reload(decal_ui)
    importlib.reload(loading_options)
    importlib.reload(admin_utils)

if "dairin0d" in locals(): importlib.reload(dairin0d)
exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "bpy_inspect": "BpyProp, prop",
    "utils_math": "matrix_LRS",
    "utils_ui": "NestedLayout, BlUI",
    "utils_blender": "BlUtil",
    "utils_view3d": "SmartView3D",
    "utils_addon": "AddonManager",
})

from . import icon_geometry
from . import common
from . import material_utils
from . import decal_utils
from . import decal_gizmos
from . import type_extensions
from . import palette
from . import palette_options
from . import palette_gizmo
from . import palette_ui
from . import template_material_menu
from . import decal_operators
from . import decal_ui
from . import loading_options
from . import admin_utils

from .common import CommonConfig, Jobs, DelayedUndoRecorder
from .material_utils import MaterialLibrary
from .decal_ui import InvertedGroupFoldState
from .palette import Palette
from .palette_options import DecalsPaletteOptions, FilesPaletteOptions
from .palette_gizmo import PaletteGizmoOptions, DecalPaletteGizmo
from .loading_options import DecalLoadingOptions
from .template_material_menu import TemplateMaterialMenu
from .admin_utils import Admin

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

# Refereces to ID blocks should be safe so long as the they're affected by Undo too?
# (i.e. not stored in WindowManager, Screen and such)

#============================================================================#

@addon.PropertyGroup
class PlacementGizmoOptions:
    auto_parent: True | prop("Auto-parent", "Automatically parent decal to the object it's been snapped to")
    sync_collections: True | prop("Sync collections", "Move decal to the same collections as the projection targets")
    
    displacement_step: 0.005 | prop("Displacement step", "Displacement step", precision=4)
    
    keep_while_dragging: {'PROJECT', 'CLIP', 'CULL', 'OCCLUDE'} | prop("Projection while dragging",
        "Which projection options to keep while dragging the decal gizmos", items=[
            ('PROJECT', "Project", ""),
            ('CLIP', "Clip", ""),
            ('CULL', "Cull", ""),
            ('OCCLUDE', "Occlude", ""),
            ('INTERSECT', "Intersect", ""),
        ])
    
    rotation_snap_by_default: False | prop("Snap rotation", "Snap rotation to increments by default")
    rotation_step: 15.0 | prop("Rotation step", "")
    rotation_step_small: 5.0 | prop("Small rotation step", "")

@addon.Internal.Include
class InternalPG:
    decal_alignment_mode: 'INITIAL' | prop("Decal alignment", "Decal alignment behavior", items=[
        ('CURRENT', "Align: current", "Align from current orientation"),
        ('INITIAL', "Align: initial", "Align from initial orientation"),
        ('UPRIGHT', "Align: upright", "Align from upright orientation"),
        ('EDGE', "Align: edge", "Align to the nearest edge"),
    ])
    
    def cycle_decal_alignment_mode(self, forward):
        items = BpyProp(self, "decal_alignment_mode")["items"]
        identifiers = [item[0] for item in items]
        i = identifiers.index(self.decal_alignment_mode)
        i = (i + (1 if forward else -1)) % len(identifiers)
        self.decal_alignment_mode = identifiers[i]
    
    default_dynamic: True | prop("Dynamic", "Whether new decals should be dynamic or baked")
    default_intersect: False | prop("Intersect", "Whether new decals should have Intersect option enabled")
    
    default_scale: 1.0 | prop("Scale", "Default scale of new decals", min=0)
    default_rotation: 'VIEW' | prop("Orientation", "Default orientation of new decals", items=[
        ('GLOBAL', "Global", "Align to world", 'ORIENTATION_GLOBAL', 1),
        ('LOCAL', "Local", "Align to the active object", 'ORIENTATION_LOCAL', 2),
        ('VIEW', "View", "Align to view", 'ORIENTATION_VIEW', 3),
        ('CURSOR', "Cursor", "Align to 3D cursor", 'ORIENTATION_CURSOR', 4),
    ])
    
    use_inherit: False | prop("Inherit", "Copy properties from the active decal or material clipboard")
    inherit_options: {'SCALE', 'GEOMETRY', 'MATERIAL'} | prop("Inherit options", "Which properties to copy", items=[
        ('SCALE', "Scale", "Inherit the active decal's object scale", 'CON_SIZELIKE'),
        ('GEOMETRY', "Geometry", "Inherit the active decal's geometry parameters", 'MESH_DATA'),
        ('MATERIAL', "Material", "Inherit the active decal's material parameters", 'MATERIAL'),
        ('CLIPBOARD', "Clipboard", "Inherit the material parameters stored in the clipboard", 'PASTEDOWN'),
    ])
    
    show_settings: True | prop("Show/hide settings", "")
    
    def _show_palette_update(self, context):
        rv3d = getattr(context, "region_data", None)
        if not isinstance(rv3d, bpy.types.RegionView3D):
            rv3d = None
            area = context.area
            if area.type == 'VIEW_3D':
                region = context.region
                if region.type != 'WINDOW':
                    info = BlUI.find_ui_area(area, region_type='WINDOW')
                    region = (info["region"] if info else None)
                if region: rv3d = BlUI.rv3d_from_region(area, region)
        
        if isinstance(rv3d, bpy.types.RegionView3D):
            DecalPaletteGizmo.active_rv3d = rv3d
        
        BlUI.tag_redraw()
    
    show_palette: {'PANEL'} | prop("Palette location", "Where to show the palette items", items=[
        ('PANEL', "Panel", "Show in Blender UI", 'MENU_PANEL', 1),
        ('VIEW3D', "3D View", "Show in 3D view", 'VIEW3D', 2),
    ], update=_show_palette_update)
    
    def _palette_sorting_update(self, context):
        self.files.refresh(reset_icons=False)
        Palette.invalidate(reset_images=False)
    palette_sorting: 'NAME' | prop("Sorting", "How to sort items in the palette", items=[
        ('NAME', "Name", "Sort by name"),
        ('TIME', "Time", "Sort by modification time"),
    ], update=_palette_sorting_update)
    
    palette: PaletteGizmoOptions | prop()
    
    placement_options: PlacementGizmoOptions | prop()
    
    decals: DecalsPaletteOptions | prop()
    
    files: FilesPaletteOptions | prop()
    
    loading_options: DecalLoadingOptions | prop()
    
    def get_default_matrix(self, context):
        scene = context.scene
        L, R, S = scene.cursor.matrix.decompose()
        S = Vector.Fill(3, self.default_scale)
        if self.default_rotation == 'GLOBAL':
            R = Quaternion()
        elif self.default_rotation == 'LOCAL':
            obj = context.object
            R = (obj.matrix_world.decompose()[1] if obj else Quaternion())
        elif self.default_rotation == 'VIEW':
            sv = SmartView3D(context, use_camera_axes=True)
            if not sv:
                area_type = (context.area if context.area.type == 'VIEW_3D' else 'VIEW_3D')
                kwargs = BlUI.find_ui_area(area_type, region_type='WINDOW')
                sv = SmartView3D(context, use_camera_axes=True, **kwargs)
            if sv: R = sv.rotation
        return matrix_LRS(L, R, S)
    
    def draw_inherit_options(self, layout):
        layout = NestedLayout(layout)
        with layout.row():
            layout.prop(self, "use_inherit", icon='DUPLICATE')
            with layout.row(align=True):
                layout.prop_enum(self, "inherit_options", 'SCALE', text="")
                layout.prop_enum(self, "inherit_options", 'GEOMETRY', text="")
                layout.prop_enum(self, "inherit_options", 'MATERIAL', text="")
                layout.prop_enum(self, "inherit_options", 'CLIPBOARD', text="")

Palette.multiprop(InternalPG, "icon_scale", 2.0 | prop("Icon scale", "Icon scale",
    min=1.0, max=10.0, step=int(CommonConfig.palette_zoom_step*100), options={'TEXTEDIT_UPDATE'}))
Palette.multiprop(InternalPG, "files_show", True | prop("", "Switch between palette and file browser modes",
    update=(lambda self, context: self.files.refresh(reset_icons=False))))

@addon.Panel(idname="VIEW3D_PT_xdecal_shortcut_config", label="Shortcut", space_type='CONSOLE', region_type='WINDOW')
class ShortcutConfigPanel:
    def draw(self, context):
        layout = NestedLayout(self.layout)
        
        layout.ui_units_x = (250 / BlUI.ICON_SIZE)
        
        addon_prefs = addon.preferences
        
        metadata = BlUI.get_ui_metadata(context)
        prefix = metadata["prefix"]
        
        key_name = f"{prefix}_key"
        mods_name = f"{prefix}_mods"
        
        layout.prop(addon_prefs, key_name, text="", event=True)
        
        row = layout.row(align=True)
        row.prop_enum_filtered(addon_prefs, mods_name)
    
    @classmethod
    def draw_shortcut(cls, layout, prefix):
        addon_prefs = addon.preferences
        
        key_name = f"{prefix}_key"
        mods_name = f"{prefix}_mods"
        
        key = getattr(addon_prefs, key_name)
        mods = getattr(addon_prefs, mods_name)
        
        key = layout.enum_item_name(addon_prefs, key_name, key)
        mods = [layout.enum_item_name(addon_prefs, mods_name, m) for m in mods]
        text = "+".join(sorted(mods) + [key])
        
        row = layout.row(align=True)
        BlUI.set_ui_metadata(row, {"prefix":prefix})
        row.popover(cls.bl_idname, text=text)

@addon.PropertyGroup
class AddonColorsPG:
    theme = None
    
    @classmethod
    def register(cls, **kwargs):
        annotations = getattr(cls, "__annotations__", None)
        if annotations is None:
            annotations = {}
            cls.__annotations__ = annotations
        
        for key, info in kwargs.items():
            value, path, name, tooltip = info
            if path:
                use = path.startswith("#")
                path = path.strip("#")
                setattr(cls, key+"_path", path.split("."))
                annotations[key+"_use"] = use | prop(name, f"Use custom {tooltip} color")
            annotations[key+"_color"] = value | prop(name, f"Custom {tooltip} color", subtype='COLOR')
    
    @classmethod
    def get_theme(cls):
        themes = bpy.context.preferences.themes
        cls.theme = themes.get("Default") or themes[0]
        return cls.theme
    
    def __getitem__(self, key):
        path = getattr(self, key+"_path", None)
        
        if (not path) or getattr(self, key+"_use", None):
            value = getattr(self, key+"_color")
        else:
            value = AddonColorsPG.theme or AddonColorsPG.get_theme()
            for part in path:
                value = getattr(value, part)
        
        value = Vector(value).to_4d()
        
        if len(value) < 4:
            value = value.to_4d()
            value[3] = 1.0
        
        return value
    
    def draw(self, layout, key):
        layout = NestedLayout(layout)
        
        with layout.split(factor=0.8, align=True):
            if hasattr(self, key+"_use"):
                layout.prop(self, key+"_use", toggle=True)
            else:
                layout.label(text=self.bl_rna.properties[key+"_color"].name)
            
            layout.prop(self, key+"_color", text="")

AddonColorsPG.register(
    decal_main=[(1.0, 0.5, 0.0, 0.75), "view_3d.object_active", "Color", "gizmo"],
    decal_highlight=[(1.0, 0.75, 0.5, 0.75), "#view_3d.object_active", "Highlight", "gizmo highlight"],
    decal_active=[(1.0, 1.0, 1.0, 0.75), "#view_3d.object_active", "Active", "active gizmo"],
    decal_edge=[(1.0, 1.0, 0.0, 1.0), "#view_3d.wire", "Snap edge", "nearest edge"],
    decal_volume=[(1.0, 1.0, 1.0, 0.1), "view_3d.editmesh_active", "Volume", "projection volume"],
    
    scroll_outline=[(0.259, 0.259, 0.259, 1.0), "user_interface.wcol_scroll.outline", "Outline", "scroll outline"],
    scroll_handle=[(0.404, 0.404, 0.404, 1.0), "user_interface.wcol_scroll.item", "Handle", "scroll handle"],
    scroll_area=[(0.2, 0.2, 0.2, 0.0), "user_interface.wcol_scroll.inner", "Background", "scroll area"],
    palette_panel=[(0.2, 0.2, 0.2, 0.941), "view_3d.space.panelcolors.back", "Background", "palette background"],
    item_selection=[(0.0, 0.5, 1.0, 1.0), "#user_interface.wcol_tool.inner_sel", "Selection", "item selection"],
    label_text=[(1.0, 1.0, 1.0, 1.0), "user_interface.wcol_text.text", "Text", "text"],
    label_shadow=[(0.0, 0.0, 0.0, 1.0), "user_interface.wcol_text.inner", "Shadow", "text shadow"],
    toolbar_widget=[(1.0, 1.0, 1.0, 1.0), "user_interface.wcol_regular.text_sel", "Widgets", "toolbar widgets"],
    toolbar_toggle=[(0.337, 0.502, 0.761, 0.902), "user_interface.wcol_regular.inner_sel", "Toggle", "toolbar toggles"],
    toolbar_active=[(0.0, 0.0, 0.0, 0.5), "#user_interface.wcol_regular.inner", "Highlight", "widget highlight"],
)

dclick_action_items = [
    ('NONE', "<No action>", "Do nothing"),
    ('SELECT', "Select", "File mode: select related files, Palette mode: select template instances"),
    ('ADD_REPLACE', "Add/Replace", "If no decals are selected: add decal(s) at cursor. Otherwise, replace them"),
    ('ADD_MODIFY', "Add/Modify", "If no decals are selected: add decal(s) at cursor. Otherwise, replace their images or materials"),
]

@addon.Preferences.Include
class DecalAddonPreferences:
    prefs_tab: 'UI' | prop("Tab", "Which tab to show in addon preferences", items=[
        ('UI', "UI / Defaults", "UI settings and various defaults"),
        ('GIZMO', "Gizmo", "Decal gizmo options"),
        ('VIEWPORT_PANEL', "3D View Palette", "3D view palette options"),
        ('ABOUT', "About", "About"),
    ])
    
    show_in_sidebar: True | prop("Show in sidebar", f"Show '{CommonConfig.sidebar_category}' tab in the 3D view's sidebar")
    show_in_header: True | prop("Show in header", f"Show an icon in the 3D view's header")
    
    split_palette_options: True | prop("Split options",
        "Use independent options in 3D View and Blender UI palettes.\n" +
        "What can be independent: file browser mode, icon scale, template filters")
    
    palette_dclick_action: 'ADD_REPLACE' | prop("Double-click", "What action to perform on palette items",
        items=dclick_action_items)
    palette_ctrl_dclick_action: 'SELECT' | prop("Ctrl+Double-click", "What action to perform on palette items",
        items=dclick_action_items)
    palette_alt_dclick_action: 'ADD_MODIFY' | prop("Alt+Double-click", "What action to perform on palette items",
        items=dclick_action_items)
    
    background_job_duration: 0.01 | prop("Icon loading", "How much time per update the addon can spend on loading previews",
        subtype='TIME', unit='TIME', min=0.0, max=0.25)
    
    preview_render_size: 128 | prop("Thumbnail size", "Resolution of decal template previews (affects file size)",
        subtype='PIXEL', min=32, max=256)
    
    lines_per_page: 0 | prop("Page size", "If > 0, palette preview will be split into pages.\nMeasured in UI lines/rows", min=0, max=50)
    
    popup_dialog_width: 300 | prop("Popup width", "Width of the popup panels/dialogs", min=100, subtype='PIXEL')
    
    popup_context_use: True | addon.prop_keymap("Use Decal Popup shortcut", "Enable a dedicated shortcut for Decal Popup")
    popup_context_key: 'RIGHTMOUSE' | addon.prop_keymap_key("Decal Popup")
    popup_context_mods: {'ctrl'} | addon.prop_keymap_mods("Decal Popup")
    
    popup_palette_use: True | addon.prop_keymap("Use Popup Palette", "Enable Popup Palette")
    popup_palette_key: 'D' | addon.prop_keymap_key("Popup Palette")
    popup_palette_mods: {'ctrl', 'alt'} | addon.prop_keymap_mods("Popup Palette")
    
    toggle_palette_use: True | addon.prop_keymap("3D View decal palette on/off", "Enable a shortcut for showing/hiding the 3D View decal palette")
    toggle_palette_key: 'D' | addon.prop_keymap_key("3D View decal palette")
    toggle_palette_mods: {'ctrl'} | addon.prop_keymap_mods("3D View decal palette")
    
    path_mode: 'RELATIVE' | prop("Convert paths", "Convert paths to absolute or relative style on import", items=[
        ('DONT_CHANGE', "Don't change", "Don't change file paths"),
        ('ABSOLUTE', "Absolute", "Make paths absolute"),
        ('RELATIVE', "Relative", "Make paths relative"),
    ])
    
    template_material_path: f"{MaterialLibrary.default}/Basic - Diffuse" | prop("Default template material", "")
    
    use_setup_snapping: False | prop("Snapping", "Configure Blender's snapping settings when a decal is created:\n"+
        "Transform Pivot Point: Bounding Box Center\n"+
        "Snap With: Center\n"+
        "Snap To: Face\n"+
        "Align Rotation to Target: enabled")
    
    placement_options: PlacementGizmoOptions | prop()
    
    colors: AddonColorsPG | prop()
    
    gizmo_anim_speed: 1.0 | prop("Gizmo animation speed", "Gizmo animation speed")
    gizmo_scale: 0.05 | prop("Gizmo icon scale", "Gizmo icon scale")
    
    auto_hide_mode: 'STATIC' | prop("Auto-hide mode", "Behavior of the gizmo palette's auto-hide", items=[
        ('STATIC', "Static", "Palette is visible when mouse is over it, and hidden otherwise"),
        ('STICKY', "Sticky", "Palette appears when mouse enters the mini-icon, and disappears when mouse leaves the palette"),
    ])
    
    palette_show_all_names: False | prop("Show all names", "Always show item names (instead of just for the highlighted item)")
    
    palette_padding: 4 | prop("Padding", min=2, max=32, subtype='PIXEL')
    palette_scroll_size: 8 | prop("Scroll size", min=2, max=32, subtype='PIXEL')
    
    palette_text_size: 12 | prop("Item text size", min=5, max=50)
    
    palette_preview_size: 128 | prop("Preview size", "The size of the enlarged item preview", min=32, subtype='PIXEL')
    palette_trigger_size: 64 | prop("Trigger size", "The size of pallete items below which an enlarged preview will be displayed", min=0, subtype='PIXEL')
    palette_preview_alpha: 0.5 | prop("Preview opacity", "Opacity of the enlarged item preview", min=0.0, max=1.0)
    
    toolbar_scale: 1.0 | prop("Toolbar scale", "Toolbar scale", min=0.0, max=5.0)
    toolbar_background_alpha: 0.5 | prop("Toolbar opacity", "Opacity of the toolbar background", min=0.0, max=1.0)
    
    def draw(self, context):
        layout = NestedLayout(self.layout)
        
        layout.row().prop_tabs_enum(self, "prefs_tab")
        
        if self.prefs_tab == 'UI':
            self.draw_settings_ui(layout)
        elif self.prefs_tab == 'GIZMO':
            self.draw_settings_gizmo(layout)
        elif self.prefs_tab == 'VIEWPORT_PANEL':
            self.draw_settings_viewport(layout)
        elif self.prefs_tab == 'ABOUT':
            self.draw_about(layout)
    
    def draw_about(self, layout):
        row = layout.row()
        col = row.column()
        col.label(text="Official:")
        col.operator("wm.url_open", text="MOTH3R® Store").url = "store.moth3r.com"
        col.operator("wm.url_open", text="MOTH3R® Docs").url = "docs.moth3r.com"
        col = row.column()
        col.label(text="Recommended:")
        col.operator("wm.url_open", text="MasterXeon1001 addons").url = "https://gumroad.com/masterxeon1001"
        col.operator("wm.url_open", text="MACHIN3 tools").url = "https://machin3.io/"
    
    def draw_settings_ui(self, layout):
        def draw_dclick_action(layout, name, text):
            value = getattr(self, name)
            label = layout.enum_item_name(self, name, value)
            layout.prop_menu_enum(self, name, text=f"{text}: {label}")
        
        with layout.split(factor=0.5):
            with layout.column():
                with layout.box():
                    with layout.row():
                        layout.label(text="Panels:")
                        layout.prop(self, "lines_per_page", text="Page size")
                    with layout.row():
                        layout.prop(self, "show_in_sidebar", text="Sidebar", toggle=True)
                        layout.prop(self, "show_in_header", text="Header", toggle=True)
                        layout.prop(self, "split_palette_options", text="Independent", toggle=True)
                    with layout.column(align=True):
                        draw_dclick_action(layout, "palette_dclick_action", "Double-click")
                        draw_dclick_action(layout, "palette_ctrl_dclick_action", "Ctrl+Double-click")
                        draw_dclick_action(layout, "palette_alt_dclick_action", "Alt+Double-click")
                with layout.box():
                    layout.label(text="Shortcuts:")
                    with layout.row():
                        layout.prop(self, "popup_context_use", text="Decal Popup")
                        with layout.row()(active=self.popup_context_use):
                            ShortcutConfigPanel.draw_shortcut(layout, "popup_context")
                    with layout.row():
                        layout.prop(self, "popup_palette_use", text="Popup Palette")
                        with layout.row()(active=self.popup_palette_use):
                            ShortcutConfigPanel.draw_shortcut(layout, "popup_palette")
                    with layout.row():
                        layout.prop(self, "toggle_palette_use", text="3D View palette")
                        with layout.row()(active=self.toggle_palette_use):
                            ShortcutConfigPanel.draw_shortcut(layout, "toggle_palette")
            with layout.column():
                with layout.box():
                    layout.label(text="UI:")
                    with layout.column():
                        layout.prop(self, "background_job_duration")
                        layout.prop(self, "preview_render_size")
                        layout.prop(self, "popup_dialog_width")
                with layout.box():
                    with layout.row():
                        layout.label(text="Defaults:")
                        InvertedGroupFoldState.draw_popup(layout, text="Folded groups", icon='NONE')
                    # Logically, showing directory presets here could make sense too,
                    # but it would require a separate popup implementation that would
                    # open directory browser when clicking on "+" or on entry name.
                    # palette_options.DirectoryPreset.draw_popup(layout, text="Directory presets", icon='PRESET')
                    with layout.row():
                        with layout.column()(alignment='LEFT'):
                            layout.label(text="Path:")
                            layout.label(text="Material:")
                        with layout.column():
                            layout.prop(self, "path_mode", text="")
                            TemplateMaterialMenu.draw_main(layout, True)
                    layout.prop(self, "use_setup_snapping", text="Auto configure snapping")
    
    def draw_settings_gizmo(self, layout):
        placement = self.placement_options
        with layout.grid_flow(even_columns=True, even_rows=True, align=True):
            with layout.column():
                layout.label(text="Behavior:")
                layout.prop_menu_enum(placement, "keep_while_dragging")
                layout.prop(placement, "displacement_step", text="Distance")
                with layout.row():
                    layout.prop(placement, "auto_parent")
                    layout.prop(placement, "sync_collections")
                layout.prop(placement, "rotation_snap_by_default", text="Snap rotation by default")
                layout.prop(placement, "rotation_step", text="Rotation Step")
                layout.prop(placement, "rotation_step_small", text="Rotation Substep")
            with layout.column():
                layout.label(text="Display:")
                layout.prop(self, "gizmo_anim_speed", text="Animation Speed")
                layout.prop(self, "gizmo_scale", text="Icon Scale")
                self.colors.draw(layout, "decal_main")
                self.colors.draw(layout, "decal_highlight")
                self.colors.draw(layout, "decal_active")
                self.colors.draw(layout, "decal_edge")
                self.colors.draw(layout, "decal_volume")
    
    def draw_settings_viewport(self, layout):
        with layout.row():
            with layout.column():
                layout.label(text="Scroll:")
                layout.prop(self, "palette_scroll_size", text="Size")
                self.colors.draw(layout, "scroll_outline")
                self.colors.draw(layout, "scroll_handle")
                self.colors.draw(layout, "scroll_area")
            
            with layout.column():
                with layout.row()(alignment='LEFT'):
                    layout.label(text="Panel:")
                    with layout.row()(alignment='EXPAND'):
                        layout.prop_menu_enum(self, "auto_hide_mode", text="Auto-hide")
                layout.prop(self, "palette_padding", text="Padding")
                
                op = layout.operator("wm.call_panel", text="Anchors", icon='FULLSCREEN_ENTER')
                op.name = "VIEW3D_PT_xdecal_palette_gizmo_anchors"
                
                self.colors.draw(layout, "palette_panel")
                self.colors.draw(layout, "item_selection")
            
            with layout.column():
                layout.label(text="Toolbar:")
                layout.prop(self, "toolbar_scale", text="Scale")
                layout.prop(self, "toolbar_background_alpha", text="Alpha", slider=True)
                self.colors.draw(layout, "toolbar_widget")
                self.colors.draw(layout, "toolbar_toggle")
                self.colors.draw(layout, "toolbar_active")
            
            with layout.column():
                with layout.row()(alignment='LEFT'):
                    layout.label(text="Items:")
                    layout.prop(self, "palette_show_all_names", text="All names", toggle=True)
                layout.prop(self, "palette_preview_size", text="Peek size")
                layout.prop(self, "palette_trigger_size", text="Peek at")
                layout.prop(self, "palette_preview_alpha", text="Peek alpha", slider=True)
                layout.prop(self, "palette_text_size", text="Text size")
                self.colors.draw(layout, "label_text")
                self.colors.draw(layout, "label_shadow")

@addon.Operator(idname="xdecal.delayed_undo_recorder", label="Delayed Undo Recorder")
class InputMonitorOperator:
    "Record Undo after the user has finished editing a value"
    def invoke(self, context, event):
        if DelayedUndoRecorder.message:
            bpy.ops.ed.undo_push(message=DelayedUndoRecorder.message)
            DelayedUndoRecorder.message = ""
        return {'PASS_THROUGH'}

@addon.on_register_keymaps
def register_keymaps():
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon
    if not kc: return
    
    #km = kc.keymaps.new(name="Object Mode")
    km = kc.keymaps.new(name="3D View", space_type='VIEW_3D')
    
    if settings.popup_context_use:
        km_kwargs = {mod:True for mod in settings.popup_context_mods}
        kmi = km.keymap_items.new('wm.call_panel', settings.popup_context_key, 'PRESS', **km_kwargs)
        kmi.properties.name = "EMPTY_PT_xdecal_context_view3d"
    
    if settings.popup_palette_use:
        km_kwargs = {mod:True for mod in settings.popup_palette_mods}
        kmi = km.keymap_items.new('xdecal.popup_palette', settings.popup_palette_key, 'PRESS', **km_kwargs)
    
    if settings.toggle_palette_use:
        km_kwargs = {mod:True for mod in settings.toggle_palette_mods}
        kmi = km.keymap_items.new('xdecal.toggle_palette_gizmo', settings.toggle_palette_key, 'PRESS', **km_kwargs)
    
    km = kc.keymaps.new(name="Window", space_type='EMPTY')
    kmi = km.keymap_items.new('xdecal.delayed_undo_recorder', 'MOUSEMOVE', 'ANY')

def register():
    Admin.reset_addon()
    addon.register()
    BlUtil.Depsgraph.trigger_update()
    

def unregister():
    Admin.reset_addon()
    addon.unregister()
