# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import os

import bpy

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "bpy_inspect": "prop",
    "utils_ui": "BlUI",
    "utils_addon": "AddonManager",
})

from . import common

from .common import CommonUtils
from .material_utils import MaterialLibrary

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

@addon.Menu(idname="SCENE_MT_xdecal_template_material", label="Template material", description="The material on which new decals will be based")
class TemplateMaterialMenu:
    @classmethod
    def draw_main(cls, layout, for_prefs=False):
        if for_prefs:
            active_path = settings.template_material_path
        else:
            active_path = settings.loading_options.template_material_path
        
        library_path, material_name = CommonUtils.split(active_path)
        
        # For some reason, the "alert" background does not display in boxes
        is_missing = not MaterialLibrary.availability.get(active_path, True)
        icon = ('ERROR' if is_missing else 'NONE')
        
        row = layout.row(align=True)
        BlUI.set_ui_metadata(row, {"for_prefs":for_prefs})
        row.menu(cls.bl_idname, text=material_name, icon=icon, translate=False)
    
    def draw(self, context):
        layout = self.layout
        
        material_infos = MaterialLibrary.read()
        
        if not material_infos:
            layout.label(text="No template materials were found!", icon='ERROR')
            return
        
        metadata = BlUI.get_ui_metadata(context)
        
        for_prefs = (metadata.get("for_prefs") == "True")
        
        if for_prefs:
            active_path = settings.template_material_path
        else:
            active_path = settings.loading_options.template_material_path
        
        prev_path = None
        for material_info in material_infos:
            material_path, abs_path, material_name = material_info
            
            if abs_path != prev_path:
                prev_path = abs_path
                library_name = os.path.basename(abs_path)
                layout.label(text=library_name, icon='FILE_BLEND', translate=False)
            
            #icon='NODE_MATERIAL' # MATERIAL
            #icon = ('CHECKBOX_HLT' if material_path == active_path else 'CHECKBOX_DEHLT')
            icon = ('RADIOBUT_ON' if material_path == active_path else 'RADIOBUT_OFF')
            op = layout.operator("xdecal.set_template_material", text=material_name, icon=icon, translate=False)
            op.path = material_path
            op.for_prefs = for_prefs

@addon.Operator(idname="xdecal.set_template_material", label="Set template material", description="Select this material")
class SetTemplateMaterialOperator:
    path: "" | prop()
    for_prefs: False | prop()
    def execute(self, context):
        if self.for_prefs:
            settings.template_material_path = self.path
        else:
            settings.loading_options.template_material_path = self.path
        return {'FINISHED'}
