# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import math
import itertools
import re

import bpy
import bmesh
from mathutils import Vector, Matrix, Quaternion, Euler, Color

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "bounds": "RangeAggregator",
    "bpy_inspect": "BlRna, prop",
    "utils_blender": "BlUtil, MeshEquivalent, Raycaster",
    "utils_addon": "AddonManager",
})

from . import common
from . import material_utils
from . import decal_utils
from . import decal_gizmos

from .common import Jobs
from .material_utils import MaterialUtils, ImageUtils
from .decal_utils import DecalUtils
from .decal_gizmos import decal_gizmo_state, GizmoShape

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

@addon.type_extension(bpy.types.Image, "xdecal")
@addon.PropertyGroup
class ImageDecalInfo:
    source_image: bpy.types.Image | prop("Source image", "Source image")
    
    def normal_map_update(self, context):
        self.make_normalmap()
    normal_map: bpy.types.Image | prop("Normal map", "Auto-generated normal map")
    normal_map_blur_type: 'BOX' | prop("Blur type", "Blur type", update=normal_map_update, items=[
        ('BOX', "Box", "Box blur", 'NOCURVE', 0),
        ('GAUSSIAN', "Gaussian", "Gaussian blur", 'SMOOTHCURVE', 1),
    ])
    normal_map_blur: 0.0 | prop("Blur", "Blur", min=0, update=normal_map_update)
    normal_map_scale: 255.0 | prop("Scale", "Scale", update=normal_map_update)
    
    def make_normalmap(self):
        image = self.id_data
        scale = self.normal_map_scale
        blur = self.normal_map_blur
        blur_type = self.normal_map_blur_type
        self.normal_map = ImageUtils.heightmap_to_normalmap(image, self.normal_map, scale, blur, blur_type)
        self.normal_map.xdecal.source_image = image
        return self.normal_map
